"""Test OOB Manager."""
import asyncio
import json
from uuid import UUID

from asynctest import mock as async_mock, TestCase as AsyncTestCase
from copy import deepcopy
from datetime import datetime, timezone
from uuid import UUID

from .....connections.models.conn_record import ConnRecord
from .....connections.models.connection_target import ConnectionTarget
from .....connections.models.diddoc import DIDDoc, PublicKey, PublicKeyType, Service
from .....core.in_memory import InMemoryProfile
from .....indy.holder import IndyHolder
from .....ledger.base import BaseLedger
from .....messaging.decorators.attach_decorator import AttachDecorator
from .....messaging.responder import BaseResponder, MockResponder
from .....messaging.util import str_to_datetime, str_to_epoch
from .....multitenant.manager import MultitenantManager
from .....protocols.connections.v1_0.manager import ConnectionManager
from .....protocols.coordinate_mediation.v1_0.models.mediation_record import (
    MediationRecord,
)
from .....protocols.coordinate_mediation.v1_0.manager import MediationManager
from .....protocols.didexchange.v1_0.manager import DIDXManager
from .....protocols.issue_credential.v1_0.message_types import (
    CREDENTIAL_OFFER,
)
from .....protocols.present_proof.v1_0.manager import PresentationManager
from .....protocols.present_proof.v1_0.message_types import (
    PRESENTATION_REQUEST,
    ATTACH_DECO_IDS,
    PRESENTATION_PREVIEW,
)
from .....protocols.present_proof.v1_0.messages.presentation import (
    Presentation,
)
from .....protocols.present_proof.v1_0.models.presentation_exchange import (
    V10PresentationExchange,
)
from .....protocols.present_proof.v1_0.message_types import PRESENTATION_REQUEST
from .....protocols.present_proof.v1_0.messages.presentation_proposal import (
    PresentationProposal,
)
from .....protocols.present_proof.v1_0.messages.presentation_request import (
    PresentationRequest,
    PresentationRequestSchema,
)
from .....protocols.present_proof.v1_0.messages.inner.presentation_preview import (
    PresAttrSpec,
    PresentationPreview,
    PresPredSpec,
)
from .....storage.error import StorageError, StorageNotFoundError
from .....multitenant.manager import MultitenantManager
from .....transport.inbound.receipt import MessageReceipt
from .....wallet.base import DIDInfo, KeyInfo
from .....wallet.in_memory import InMemoryWallet
from .....wallet.util import did_key_to_naked, naked_to_did_key

from ....didcomm_prefix import DIDCommPrefix
from ....issue_credential.v1_0.models.credential_exchange import V10CredentialExchange

from .. import manager as test_module
from ..manager import (
    OutOfBandManager,
    OutOfBandManagerError,
    OutOfBandManagerNotImplementedError,
)
from ..message_types import INVITATION
from ..messages.invitation import HSProto, InvitationMessage, InvitationMessageSchema
from ..messages.reuse import HandshakeReuse
from ..messages.reuse_accept import HandshakeReuseAccept
from ..messages.problem_report import ProblemReport, ProblemReportReason
from ..models.invitation import InvitationRecord


class TestConfig:

    test_did = "55GkHamhTU1ZbTbV2ab9DE"
    test_verkey = "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx"
    test_endpoint = "http://localhost"
    test_target_did = "GbuDUYXaUZRfHD2jeDuQuP"
    their_public_did = "55GkHamhTU1ZbTbV2ab9DE"
    NOW_8601 = datetime.utcnow().replace(tzinfo=timezone.utc).isoformat(" ", "seconds")
    NOW_EPOCH = str_to_epoch(NOW_8601)
    CD_ID = "GMm4vMw8LLrLJjp81kRRLp:3:CL:12:tag"
    INDY_PROOF_REQ = json.loads(
        f"""{{
        "name": "proof-req",
        "version": "1.0",
        "nonce": "12345",
        "requested_attributes": {{
            "0_player_uuid": {{
                "name": "player",
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }},
            "0_screencapture_uuid": {{
                "name": "screenCapture",
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }}
        }},
        "requested_predicates": {{
            "0_highscore_GE_uuid": {{
                "name": "highScore",
                "p_type": ">=",
                "p_value": 1000000,
                "restrictions": [
                    {{
                        "cred_def_id": "{CD_ID}"
                    }}
                ],
                "non_revoked": {{
                    "from": {NOW_EPOCH},
                    "to": {NOW_EPOCH}
                }}
            }}
        }}
    }}"""
    )

    PRES_PREVIEW = PresentationPreview(
        attributes=[
            PresAttrSpec(name="player", cred_def_id=CD_ID, value="Richie Knucklez"),
            PresAttrSpec(
                name="screenCapture",
                cred_def_id=CD_ID,
                mime_type="image/png",
                value="aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
            ),
        ],
        predicates=[
            PresPredSpec(
                name="highScore", cred_def_id=CD_ID, predicate=">=", threshold=1000000
            )
        ],
    )

    PRES_REQ = PresentationRequest(
        comment="Test",
        request_presentations_attach=[
            AttachDecorator.data_base64(
                mapping=INDY_PROOF_REQ,
                ident=ATTACH_DECO_IDS[PRESENTATION_REQUEST],
            )
        ],
    )

    pres_req_dict = PRES_REQ.request_presentations_attach[0].serialize()
    req_attach = {
        "@id": "request-0",
        "mime-type": "application/json",
        "data": {
            "json": {
                "@type": DIDCommPrefix.qualify_current(PRESENTATION_REQUEST),
                "@id": "12345678-1234-5678-1234-567812345678",
                "comment": "some comment",
                "request_presentations~attach": [pres_req_dict],
            }
        },
    }

    indy_cred_req = {
        "schema_id": f"{test_did}:2:bc-reg:1.0",
        "cred_def_id": f"{test_did}:3:CL:12:tag1",
    }
    cred_req_meta = {}

    def make_did_doc(self, did, verkey):
        doc = DIDDoc(did=did)
        controller = did
        ident = "1"
        pk_value = verkey
        pk = PublicKey(
            did, ident, pk_value, PublicKeyType.ED25519_SIG_2018, controller, False
        )
        doc.set(pk)
        recip_keys = [pk]
        router_keys = []
        service = Service(
            did, "indy", "IndyAgent", recip_keys, router_keys, TestConfig.test_endpoint
        )
        doc.set(service)
        return doc


class TestOOBManager(AsyncTestCase, TestConfig):
    def setUp(self):
        self.responder = MockResponder()
        self.responder.send = async_mock.CoroutineMock()

        self.session = InMemoryProfile.test_session(
            {
                "default_endpoint": TestConfig.test_endpoint,
                "default_label": "This guy",
                "additional_endpoints": ["http://aries.ca/another-endpoint"],
                "debug.auto_accept_invites": True,
                "debug.auto_accept_requests": True,
            }
        )
        self.session.context.injector.bind_instance(BaseResponder, self.responder)
        self.mt_mgr = async_mock.MagicMock()
        self.mt_mgr = async_mock.create_autospec(MultitenantManager)
        self.session.context.injector.bind_instance(MultitenantManager, self.mt_mgr)

        self.multitenant_mgr = async_mock.MagicMock(MultitenantManager, autospec=True)
        self.session.context.injector.bind_instance(
            MultitenantManager, self.multitenant_mgr
        )

        self.ledger = async_mock.create_autospec(BaseLedger)
        self.ledger.__aenter__ = async_mock.CoroutineMock(return_value=self.ledger)
        self.ledger.get_endpoint_for_did = async_mock.CoroutineMock(
            return_value=TestConfig.test_endpoint
        )
        self.session.context.injector.bind_instance(BaseLedger, self.ledger)

        self.manager = OutOfBandManager(self.session)
        assert self.manager.session

        self.test_conn_rec = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_role=None,
            state=ConnRecord.State.COMPLETED,
            their_public_did=self.their_public_did,
        )

        self.test_mediator_routing_keys = [
            "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRR"
        ]
        self.test_mediator_conn_id = "mediator-conn-id"
        self.test_mediator_endpoint = "http://mediator.example.com"

    async def test_create_invitation_handshake_succeeds(self):
        self.session.context.update_settings({"public_invites": True})

        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[HSProto.RFC23],
            )

            assert invi_rec.invitation["@type"] == DIDCommPrefix.qualify_current(
                INVITATION
            )
            assert not invi_rec.invitation.get("request~attach")
            assert (
                DIDCommPrefix.qualify_current(HSProto.RFC23.name)
                in invi_rec.invitation["handshake_protocols"]
            )
            assert invi_rec.invitation["service"] == [f"did:sov:{TestConfig.test_did}"]

    async def test_create_invitation_mediation_overwrites_routing_and_endpoint(self):
        mock_conn_rec = async_mock.MagicMock()

        mediation_record = MediationRecord(
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )
        await mediation_record.save(self.session)
        with async_mock.patch.object(
            MediationManager,
            "get_default_mediator_id",
        ) as mock_get_default_mediator, async_mock.patch.object(
            mock_conn_rec, "metadata_set", async_mock.CoroutineMock()
        ) as mock_metadata_set:
            invite = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                my_label="test123",
                hs_protos=[HSProto.RFC23],
                mediation_id=mediation_record.mediation_id,
            )
            assert isinstance(invite, InvitationRecord)
            assert invite.invitation["@type"] == DIDCommPrefix.qualify_current(
                INVITATION
            )
            assert invite.invitation["label"] == "test123"
            mock_get_default_mediator.assert_not_called()

    async def test_create_invitation_ledger_x(self):
        self.session.context.update_settings({"public_invites": True})

        self.ledger.__aenter__ = async_mock.CoroutineMock(return_value=self.ledger)
        self.session.context.injector.bind_instance(BaseLedger, self.ledger)

        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did, async_mock.patch.object(
            self.ledger, "get_endpoint_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_endpoint_for_did:
            mock_ledger_get_endpoint_for_did.side_effect = test_module.LedgerError()
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    my_endpoint=TestConfig.test_endpoint,
                    public=True,
                    hs_protos=[HSProto.RFC23],
                )
            assert "Error getting endpoint" in str(context.exception)

    async def test_create_invitation_multitenant_local(self):
        self.session.context.update_settings(
            {
                "multitenant.enabled": True,
                "wallet.id": "test_wallet",
            }
        )

        self.multitenant_mgr.add_key = async_mock.CoroutineMock()

        with async_mock.patch.object(
            InMemoryWallet, "create_signing_key", autospec=True
        ) as mock_wallet_create_signing_key, async_mock.patch.object(
            self.multitenant_mgr, "get_default_mediator"
        ) as mock_get_default_mediator:
            mock_wallet_create_signing_key.return_value = KeyInfo(
                TestConfig.test_verkey, None
            )
            mock_get_default_mediator.return_value = MediationRecord()
            await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                hs_protos=[HSProto.RFC23],
                multi_use=False,
            )

            self.multitenant_mgr.add_key.assert_called_once_with(
                "test_wallet", TestConfig.test_verkey
            )

    async def test_create_invitation_multitenant_public(self):
        self.session.context.update_settings(
            {
                "multitenant.enabled": True,
                "wallet.id": "test_wallet",
                "public_invites": True,
            }
        )

        self.multitenant_mgr.add_key = async_mock.CoroutineMock()

        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                self.test_did, self.test_verkey, None
            )
            await self.manager.create_invitation(
                hs_protos=[HSProto.RFC23],
                public=True,
                multi_use=False,
            )

            self.multitenant_mgr.add_key.assert_called_once_with(
                "test_wallet", TestConfig.test_verkey, skip_if_exists=True
            )

    async def test_create_invitation_no_handshake_no_attachments_x(self):
        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=None,
                multi_use=False,
            )
            assert "Invitation must include" in str(context.exception)

    async def test_create_invitation_attachment_v1_0_cred_offer(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did, async_mock.patch.object(
            V10CredentialExchange,
            "retrieve_by_id",
            async_mock.CoroutineMock(),
        ) as mock_retrieve_cxid:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            mock_retrieve_cxid.return_value = async_mock.MagicMock(
                credential_offer_dict={"cred": "offer"}
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[HSProto.RFC23],
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            assert isinstance(invi_rec, InvitationRecord)

    async def test_create_invitation_attachment_v1_0_cred_offer_no_handshake(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did, async_mock.patch.object(
            V10CredentialExchange,
            "retrieve_by_id",
            async_mock.CoroutineMock(),
        ) as mock_retrieve_cxid:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            mock_retrieve_cxid.return_value = async_mock.MagicMock(
                credential_offer_dict={"cred": "offer"}
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=None,
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            assert isinstance(invi_rec, InvitationRecord)
            assert not invi_rec.invitation["handshake_protocols"]

    async def test_create_invitation_attachment_v2_0_cred_offer(self):
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did, async_mock.patch.object(
            test_module.V20CredOffer, "deserialize", autospec=True
        ) as mock_v20_cred_offer_deser, async_mock.patch.object(
            test_module.V10CredentialExchange,
            "retrieve_by_id",
            async_mock.CoroutineMock(),
        ) as mock_retrieve_cxid_v1, async_mock.patch.object(
            test_module.V20CredExRecord,
            "retrieve_by_id",
            async_mock.CoroutineMock(),
        ) as mock_retrieve_cxid_v2:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            mock_v20_cred_offer_deser.return_value = async_mock.MagicMock(
                offer=async_mock.MagicMock(return_value={"cred": "offer"})
            )
            mock_retrieve_cxid_v1.side_effect = test_module.StorageNotFoundError()
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=False,
                hs_protos=None,
                multi_use=False,
                attachments=[{"type": "credential-offer", "id": "dummy-id"}],
            )

            assert invi_rec.invitation["request~attach"]

    async def test_create_invitation_attachment_present_proof(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did, async_mock.patch.object(
            test_module.V10PresentationExchange,
            "retrieve_by_id",
            async_mock.CoroutineMock(),
        ) as mock_retrieve_pxid:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            mock_retrieve_pxid.return_value = async_mock.MagicMock(
                presentation_request_dict={"pres": "req"}
            )
            invi_rec = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                public=True,
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
                attachments=[{"type": "present-proof", "id": "dummy-id"}],
            )

            assert invi_rec.invitation["request~attach"]
            mock_retrieve_pxid.assert_called_once_with(self.manager.session, "dummy-id")

    async def test_create_invitation_public_x_no_public_invites(self):
        self.session.context.update_settings({"public_invites": False})

        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                public=True,
                my_endpoint="testendpoint",
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
            )
        assert "Public invitations are not enabled" in str(context.exception)

    async def test_create_invitation_public_x_multi_use(self):
        self.session.context.update_settings({"public_invites": True})

        with self.assertRaises(OutOfBandManagerError) as context:
            await self.manager.create_invitation(
                public=True,
                my_endpoint="testendpoint",
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=True,
            )
        assert "Cannot create public invitation with" in str(context.exception)

    async def test_create_invitation_public_x_no_public_did(self):
        self.session.context.update_settings({"public_invites": True})

        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = None
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    public=True,
                    my_endpoint="testendpoint",
                    hs_protos=[test_module.HSProto.RFC23],
                    multi_use=False,
                )
        assert "Cannot create public invitation with no public DID" in str(
            context.exception
        )

    async def test_create_invitation_attachment_x(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    my_endpoint=TestConfig.test_endpoint,
                    public=False,
                    hs_protos=[test_module.HSProto.RFC23],
                    multi_use=True,
                    attachments=[{"having": "attachment", "is": "no", "good": "here"}],
                )
            assert "Unknown attachment type" in str(context.exception)

    async def test_create_invitation_peer_did(self):
        self.session.context.update_settings(
            {
                "multitenant.enabled": True,
                "wallet.id": "my-wallet",
            }
        )
        mediation_record = MediationRecord(
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )
        await mediation_record.save(self.session)
        with async_mock.patch.object(
            self.multitenant_mgr, "get_default_mediator"
        ) as mock_get_default_mediator:
            mock_get_default_mediator.return_value = mediation_record
            invi_rec = await self.manager.create_invitation(
                my_label="That guy",
                my_endpoint=None,
                public=False,
                hs_protos=[test_module.HSProto.RFC23],
                multi_use=False,
            )

            assert invi_rec.invitation["@type"] == DIDCommPrefix.qualify_current(
                INVITATION
            )
            assert not invi_rec.invitation.get("request~attach")
            assert invi_rec.invitation["label"] == "That guy"
            assert (
                DIDCommPrefix.qualify_current(HSProto.RFC23.name)
                in invi_rec.invitation["handshake_protocols"]
            )
            service = invi_rec.invitation["service"][0]
            assert service["id"] == "#inline"
            assert service["type"] == "did-communication"
            assert len(service["recipientKeys"]) == 1
            assert service["routingKeys"][0] == naked_to_did_key(
                self.test_mediator_routing_keys[0]
            )
            assert service["serviceEndpoint"] == self.test_mediator_endpoint

    async def test_create_invitation_metadata_assigned(self):
        invi_rec = await self.manager.create_invitation(
            hs_protos=[test_module.HSProto.RFC23],
            metadata={"hello": "world"},
        )
        service = invi_rec.invitation["service"][0]
        invitation_key = did_key_to_naked(service["recipientKeys"][0])
        record = await ConnRecord.retrieve_by_invitation_key(
            self.session, invitation_key
        )
        assert await record.metadata_get_all(self.session) == {"hello": "world"}

    async def test_create_invitation_x_public_metadata(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            InMemoryWallet, "get_public_did", autospec=True
        ) as mock_wallet_get_public_did:
            mock_wallet_get_public_did.return_value = DIDInfo(
                TestConfig.test_did, TestConfig.test_verkey, None
            )
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_invitation(
                    public=True,
                    hs_protos=[test_module.HSProto.RFC23],
                    metadata={"hello": "world"},
                    multi_use=False,
                )
            assert "Cannot store metadata on public" in str(context.exception)

    async def test_receive_invitation_with_valid_mediation(self):
        self.session.context.update_settings({"public_invites": True})
        mediation_record = MediationRecord(
            role=MediationRecord.ROLE_CLIENT,
            state=MediationRecord.STATE_GRANTED,
            connection_id=self.test_mediator_conn_id,
            routing_keys=self.test_mediator_routing_keys,
            endpoint=self.test_mediator_endpoint,
        )
        await mediation_record.save(self.session)
        with async_mock.patch.object(
            DIDXManager, "receive_invitation", async_mock.CoroutineMock()
        ) as mock_didx_recv_invi:
            invite = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                my_label="test123",
                hs_protos=[HSProto.RFC23],
            )
            invi_msg = InvitationMessage.deserialize(invite.invitation)
            invitee_record = await self.manager.receive_invitation(
                invi_msg=invi_msg,
                mediation_id=mediation_record._id,
            )
            mock_didx_recv_invi.assert_called_once_with(
                invitation=invi_msg,
                their_public_did=None,
                auto_accept=None,
                alias=None,
                mediation_id=mediation_record._id,
            )

    async def test_receive_invitation_with_invalid_mediation(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            DIDXManager,
            "receive_invitation",
            async_mock.CoroutineMock(),
        ) as mock_didx_recv_invi:
            invite = await self.manager.create_invitation(
                my_endpoint=TestConfig.test_endpoint,
                my_label="test123",
                hs_protos=[HSProto.RFC23],
            )
            invi_msg = InvitationMessage.deserialize(invite.invitation)
            invitee_record = await self.manager.receive_invitation(
                invi_msg,
                mediation_id="test-mediation-id",
            )
            mock_didx_recv_invi.assert_called_once_with(
                invitation=invi_msg,
                their_public_did=None,
                auto_accept=None,
                alias=None,
                mediation_id=None,
            )

    async def test_receive_invitation_didx_service_block(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            test_module, "DIDXManager", autospec=True
        ) as didx_mgr_cls, async_mock.patch.object(
            test_module,
            "InvitationMessage",
            autospec=True,
        ) as invi_msg_cls:
            didx_mgr_cls.return_value = async_mock.MagicMock(
                receive_invitation=async_mock.CoroutineMock()
            )
            mock_oob_invi = async_mock.MagicMock(
                request_attach=[],
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[],
                service_blocks=[
                    async_mock.MagicMock(
                        recipient_keys=["dummy"],
                        routing_keys=[],
                    )
                ],
            )
            invi_msg_cls.deserialize.return_value = mock_oob_invi

            await self.manager.receive_invitation(mock_oob_invi)

    async def test_receive_invitation_connection_mock(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            test_module, "ConnectionManager", autospec=True
        ) as conn_mgr_cls, async_mock.patch.object(
            test_module,
            "InvitationMessage",
            autospec=True,
        ) as invi_msg_cls, async_mock.patch.object(
            self.manager,
            "receive_invitation",
            async_mock.CoroutineMock(),
        ) as mock_receive_invitation:
            mock_receive_invitation.return_value = self.test_conn_rec.serialize()
            conn_mgr_cls.return_value = async_mock.MagicMock(
                receive_invitation=async_mock.CoroutineMock()
            )
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC160.name) for pfx in DIDCommPrefix
                ],
                service_dids=[],
                label="test",
                _id="test123",
                service_blocks=[
                    async_mock.MagicMock(
                        recipient_keys=[
                            naked_to_did_key(
                                "9WCgWKUaAJj3VWxxtzvvMQN3AoFxoBtBDo9ntwJnVVCC"
                            )
                        ],
                        routing_keys=[],
                        service_endpoint="http://localhost",
                    )
                ],
                request_attach=[],
            )
            invi_msg_cls.deserialize.return_value = mock_oob_invi
            result = await self.manager.receive_invitation(mock_oob_invi)
            assert result == self.test_conn_rec.serialize()

    async def test_receive_invitation_connection(self):
        self.session.context.update_settings({"public_invites": True})
        oob_invi_rec = await self.manager.create_invitation(
            auto_accept=True,
            public=False,
            hs_protos=[test_module.HSProto.RFC160],
            multi_use=False,
        )

        result = await self.manager.receive_invitation(
            invi_msg=InvitationMessage.deserialize(oob_invi_rec.invitation),
            use_existing_connection=True,
            auto_accept=True,
        )
        conn_id = UUID(result.get("connection_id"), version=4)
        assert (
            conn_id.hex == result.get("connection_id").replace("-", "")
            and len(result.get("connection_id")) > 5
        )

    async def test_receive_invitation_no_service_blocks_nor_dids(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            test_module, "InvitationMessage", async_mock.MagicMock()
        ) as invi_msg_cls:
            mock_invi_msg = async_mock.MagicMock(
                service_blocks=[],
                service_dids=[],
            )
            invi_msg_cls.deserialize.return_value = mock_invi_msg
            with self.assertRaises(OutOfBandManagerError):
                await self.manager.receive_invitation(mock_invi_msg)

    async def test_receive_invitation_service_did(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            test_module, "DIDXManager", autospec=True
        ) as didx_mgr_cls, async_mock.patch.object(
            test_module, "InvitationMessage", autospec=True
        ) as invi_msg_cls:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            didx_mgr_cls.return_value = async_mock.MagicMock(
                receive_invitation=async_mock.CoroutineMock()
            )
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_did],
                service_blocks=[],
                request_attach=[],
            )
            invi_msg_cls.deserialize.return_value = mock_oob_invi

            invi_rec = await self.manager.receive_invitation(mock_oob_invi)
            assert invi_rec.invitation["service"]

    async def test_receive_invitation_attachment_x(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            mock_oob_invi = async_mock.MagicMock(
                service_blocks=[],
                service_dids=[TestConfig.test_did],
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                request_attach=[{"having": "attachment", "is": "no", "good": "here"}],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.receive_invitation(mock_oob_invi)
                assert (
                    "request~attach is not properly formatted as data is missing"
                    in str(context.exception)
                )

    async def test_receive_invitation_req_pres_attachment_x(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_did],
                service_blocks=[],
                request_attach=[
                    async_mock.MagicMock(
                        data=async_mock.MagicMock(
                            json={
                                "@type": DIDCommPrefix.qualify_current(
                                    PRESENTATION_REQUEST
                                )
                            }
                        )
                    ),
                ],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            with self.assertRaises(OutOfBandManagerError) as context:
                result = await self.manager.receive_invitation(mock_oob_invi)
                conn_id = UUID(result.get("connection_id"), version=4)
                assert (
                    conn_id.hex == result.get("connection_id")
                    and len(result.get("connection_id")) > 5
                )

    async def test_receive_invitation_invalid_request_type_x(self):
        self.session.context.update_settings({"public_invites": True})
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            mock_oob_invi = async_mock.MagicMock(
                service_blocks=[],
                service_dids=[TestConfig.test_did],
                handshake_protocols=[],
                request_attach=[],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            with self.assertRaises(OutOfBandManagerError):
                await self.manager.receive_invitation(mock_oob_invi)

    async def test_find_existing_connection(self):
        test_conn_rec = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_role=None,
            state=ConnRecord.State.COMPLETED,
            their_public_did=self.their_public_did,
        )
        await test_conn_rec.save(self.session)

        tag_filter = {}
        post_filter = {}
        post_filter["their_public_did"] = "not_addded"
        conn_record = await self.manager.find_existing_connection(
            tag_filter, post_filter
        )
        assert conn_record == None

        post_filter["their_public_did"] = self.their_public_did
        conn_record = await self.manager.find_existing_connection(
            tag_filter, post_filter
        )
        assert conn_record == test_conn_rec
        await test_conn_rec.delete_record(self.session)

    async def test_find_existing_connection_no_active(self):
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.REQUEST.rfc160
        await self.test_conn_rec.save(self.session)
        tag_filter = {}
        post_filter = {}
        post_filter["invitation_msg_id"] = "test_123"
        conn_record = await self.manager.find_existing_connection(
            tag_filter, post_filter
        )
        assert conn_record is None

    async def test_check_reuse_msg_state(self):
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "accepted"
        )
        assert await self.manager.check_reuse_msg_state(self.test_conn_rec) is None

    async def test_create_handshake_reuse_msg(self):
        self.session.context.update_settings({"public_invites": True})
        await self.test_conn_rec.save(self.session)
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_fetch_conn.return_value = ConnectionTarget(
                did=TestConfig.test_did,
                endpoint=TestConfig.test_endpoint,
                recipient_keys=TestConfig.test_verkey,
                sender_key=TestConfig.test_verkey,
            )
            oob_invi = InvitationMessage()

            await self.manager.create_handshake_reuse_message(
                oob_invi, self.test_conn_rec
            )
            assert (
                len(await self.test_conn_rec.metadata_get(self.session, "reuse_msg_id"))
                > 6
            )
            assert (
                await self.test_conn_rec.metadata_get(self.session, "reuse_msg_state")
                == "initial"
            )

    async def test_create_handshake_reuse_msg_catch_exception(self):
        self.session.context.update_settings({"public_invites": True})
        await self.test_conn_rec.save(self.session)
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_fetch_conn.side_effect = StorageNotFoundError()
            oob_invi = InvitationMessage()
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.create_handshake_reuse_message(
                    oob_invi, self.test_conn_rec
                )
                assert "Error on creating and sending a handshake reuse message" in str(
                    context.exception
                )

    async def test_receive_reuse_message_existing_found(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
        )
        reuse_msg = HandshakeReuse()
        reuse_msg.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            InvitationRecord,
            "retrieve_by_tag_filter",
            autospec=True,
        ) as retrieve_invi_rec:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = self.test_conn_rec
            oob_mgr_fetch_conn.return_value = ConnectionTarget(
                did=TestConfig.test_did,
                endpoint=TestConfig.test_endpoint,
                recipient_keys=TestConfig.test_verkey,
                sender_key=TestConfig.test_verkey,
            )
            oob_invi = InvitationMessage()
            retrieve_invi_rec.return_value = InvitationRecord(invi_msg_id="test_123")
            await self.manager.receive_reuse_message(reuse_msg, receipt)
            assert (
                len(
                    await ConnRecord.query(
                        session=self.session,
                        tag_filter={},
                        post_filter_positive={"invitation_msg_id": "test_123"},
                        alt=True,
                    )
                )
                == 1
            )

    async def test_receive_reuse_message_existing_not_found(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg = HandshakeReuse()
        reuse_msg.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.REQUEST.rfc160
        await self.test_conn_rec.save(self.session)
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            InvitationRecord,
            "retrieve_by_tag_filter",
            autospec=True,
        ) as retrieve_invi_rec, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = None
            oob_mgr_fetch_conn.return_value = ConnectionTarget(
                did=TestConfig.test_did,
                endpoint=TestConfig.test_endpoint,
                recipient_keys=TestConfig.test_verkey,
                sender_key=TestConfig.test_verkey,
            )
            oob_invi = InvitationMessage()
            retrieve_invi_rec.return_value = InvitationRecord(invi_msg_id="test_123")
            await self.manager.receive_reuse_message(reuse_msg, receipt)
            assert len(self.responder.messages) == 0

    async def test_receive_reuse_message_storage_not_found(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg = HandshakeReuse()
        reuse_msg.assign_thread_id(thid="test_123", pthid="test_123")

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            InvitationRecord,
            "retrieve_by_tag_filter",
            autospec=True,
        ) as retrieve_invi_rec, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.side_effect = StorageNotFoundError()
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.receive_reuse_message(reuse_msg, receipt)
                assert "No existing ConnRecord found for OOB Invitee" in str(
                    context.exception
                )

    async def test_receive_reuse_message_problem_report_logic(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg = HandshakeReuse()
        reuse_msg.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_456"
        self.test_conn_rec.their_did = "test_did"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_fetch_conn.return_value = ConnectionTarget(
                did=TestConfig.test_did,
                endpoint=TestConfig.test_endpoint,
                recipient_keys=TestConfig.test_verkey,
                sender_key=TestConfig.test_verkey,
            )
            await self.manager.receive_reuse_message(reuse_msg, receipt)

    async def test_receive_reuse_accepted(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(self.session, "reuse_msg_id", "test_123")
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "initial"
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            await self.manager.receive_reuse_accepted_message(
                reuse_msg_accepted, receipt, self.test_conn_rec
            )
            assert (
                await self.test_conn_rec.metadata_get(self.session, "reuse_msg_state")
                == "accepted"
            )

    async def test_receive_reuse_accepted(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(self.session, "reuse_msg_id", "test_123")
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "initial"
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            await self.manager.receive_reuse_accepted_message(
                reuse_msg_accepted, receipt, self.test_conn_rec
            )
            assert (
                await self.test_conn_rec.metadata_get(self.session, "reuse_msg_state")
                == "accepted"
            )

    async def test_receive_reuse_accepted_invalid_conn(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="test_123", pthid="test_123")
        test_invalid_conn = ConnRecord(
            my_did="Test",
            their_did="Test",
            invitation_msg_id="test_456",
            connection_id="12345678-1234-5678-1234-567812345678",
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.receive_reuse_accepted_message(
                    reuse_msg_accepted, receipt, test_invalid_conn
                )
                assert "Error processing reuse accepted message" in str(
                    context.exception
                )

    async def test_receive_reuse_accepted_message_catch_exception(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        reuse_msg_accepted = HandshakeReuseAccept()
        reuse_msg_accepted.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(self.session, "reuse_msg_id", "test_123")
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "initial"
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            self.test_conn_rec,
            "metadata_set",
            async_mock.CoroutineMock(side_effect=StorageNotFoundError),
        ):
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.receive_reuse_accepted_message(
                    reuse_msg_accepted, receipt, self.test_conn_rec
                )
                assert "Error processing reuse accepted message" in str(
                    context.exception
                )

    async def test_problem_report_received_not_active(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        problem_report = ProblemReport(
            problem_code=ProblemReportReason.EXISTING_CONNECTION_NOT_ACTIVE.value,
            explain="test",
        )
        problem_report.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(self.session, "reuse_msg_id", "test_123")
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "initial"
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            await self.manager.receive_problem_report(
                problem_report, receipt, self.test_conn_rec
            )
            assert (
                await self.test_conn_rec.metadata_get(self.session, "reuse_msg_state")
                == "not_accepted"
            )

    async def test_problem_report_received_not_exists(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        problem_report = ProblemReport(
            problem_code=ProblemReportReason.EXISTING_CONNECTION_DOES_NOT_EXISTS.value,
            explain="test",
        )
        problem_report.assign_thread_id(thid="test_123", pthid="test_123")
        self.test_conn_rec.invitation_msg_id = "test_123"
        self.test_conn_rec.state = ConnRecord.State.COMPLETED.rfc160
        await self.test_conn_rec.save(self.session)
        await self.test_conn_rec.metadata_set(self.session, "reuse_msg_id", "test_123")
        await self.test_conn_rec.metadata_set(
            self.session, "reuse_msg_state", "initial"
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            await self.manager.receive_problem_report(
                problem_report, receipt, self.test_conn_rec
            )
            assert (
                await self.test_conn_rec.metadata_get(self.session, "reuse_msg_state")
                == "not_accepted"
            )

    async def test_problem_report_received_invalid_conn(self):
        self.session.context.update_settings({"public_invites": True})
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did="test_did",
        )
        problem_report = ProblemReport(
            problem_code=ProblemReportReason.EXISTING_CONNECTION_DOES_NOT_EXISTS.value,
            explain="test",
        )
        problem_report.assign_thread_id(thid="test_123", pthid="test_123")
        test_invalid_conn = ConnRecord(
            my_did="Test",
            their_did="Test",
            invitation_msg_id="test_456",
            connection_id="12345678-1234-5678-1234-567812345678",
        )
        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey

            with self.assertRaises(OutOfBandManagerError) as context:
                await self.manager.receive_problem_report(
                    problem_report, receipt, test_invalid_conn
                )
                assert "Error processing problem report message" in str(
                    context.exception
                )

    async def test_existing_conn_record_public_did(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            oob_mgr_check_reuse_state.return_value = None
            oob_mgr_create_reuse_msg.return_value = None
            oob_mgr_receive_reuse_msg.return_value = None
            oob_mgr_receive_accept_msg.return_value = None
            oob_mgr_receive_problem_report.return_value = None
            await test_exist_conn.metadata_set(
                self.session, "reuse_msg_state", "accepted"
            )
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=True
            )
            retrieved_conn_records = await ConnRecord.query(
                session=self.session,
                tag_filter={},
                post_filter_positive={
                    "invitation_msg_id": "12345678-1234-5678-1234-567812345678"
                },
                alt=True,
            )
            assert (
                await retrieved_conn_records[0].metadata_get(
                    self.session, "reuse_msg_id"
                )
                is None
            )
            assert (
                await retrieved_conn_records[0].metadata_get(
                    self.session, "reuse_msg_state"
                )
                is None
            )
            assert (
                result.get("connection_id") == retrieved_conn_records[0].connection_id
            )

    async def test_existing_conn_record_public_did_not_accepted(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did="did:sov:LjgpST2rjsoxYegQDRm7EL",
            their_public_did="did:sov:LjgpST2rjsoxYegQDRm7EL",
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")

        test_new_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did="did:sov:LjgpST2rjsoxYegQDRm7EL",
            their_public_did="did:sov:LjgpST2rjsoxYegQDRm7EL",
            invitation_msg_id="12345678-1234-5678-1234-1234545454487",
            their_role=ConnRecord.Role.REQUESTER,
        )

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            oob_mgr_check_reuse_state.return_value = None
            oob_mgr_create_reuse_msg.return_value = None
            oob_mgr_receive_reuse_msg.return_value = None
            oob_mgr_receive_accept_msg.return_value = None
            oob_mgr_receive_problem_report.return_value = None
            await test_exist_conn.metadata_set(
                self.session, "reuse_msg_state", "not_accepted"
            )
            didx_mgr_receive_invitation.return_value = test_new_conn
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=True
            )
            retrieved_conn_records = await ConnRecord.query(
                session=self.session,
                tag_filter={},
                post_filter_positive={
                    "invitation_msg_id": "12345678-1234-5678-1234-567812345678"
                },
                alt=True,
            )
            assert (
                await retrieved_conn_records[0].metadata_get(
                    self.session, "reuse_msg_state"
                )
                == "not_accepted"
            )
            assert (
                result.get("connection_id") != retrieved_conn_records[0].connection_id
            )

    async def test_existing_conn_record_public_did_inverse_cases(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await self.test_conn_rec.save(self.session)
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            didx_mgr_receive_invitation.return_value = self.test_conn_rec
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=False
            )
            retrieved_conn_records = await ConnRecord.query(
                session=self.session,
                tag_filter={},
                post_filter_positive={
                    "invitation_msg_id": "12345678-1234-5678-1234-567812345678"
                },
                alt=True,
            )
            assert (
                result.get("connection_id") != retrieved_conn_records[0].connection_id
            )

    async def test_existing_conn_record_public_did_timeout(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            oob_mgr_check_reuse_state.side_effect = asyncio.TimeoutError
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi

            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=True
            )
            retrieved_conn_records = await ConnRecord.query(
                session=self.session,
                tag_filter={},
                post_filter_positive={"their_public_did": TestConfig.test_target_did},
                alt=True,
            )
            assert retrieved_conn_records[0].state == ConnRecord.State.ABANDONED.rfc160

    async def test_existing_conn_record_public_did_timeout_no_handshake_protocol(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[{"having": "attachment", "is": "no", "good": "here"}],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi
            with self.assertRaises(OutOfBandManagerError) as context:
                result = await self.manager.receive_invitation(
                    mock_oob_invi, use_existing_connection=False
                )
                assert "No existing connection exists and " in str(context.exception)

    async def test_req_attach_presentation_existing_conn_no_auto_present(self):
        self.session.context.update_settings({"public_invites": True})
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        exchange_rec = V10PresentationExchange()

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager, "receive_invitation", autospec=True
        ) as didx_mgr_receive_invitation, async_mock.patch.object(
            PresentationManager, "receive_request", autospec=True
        ) as proof_mgr_receive_request, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            proof_mgr_receive_request.return_value = exchange_rec

            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[AttachDecorator.deserialize(TestConfig.req_attach)],
            )

            inv_message_cls.deserialize.return_value = mock_oob_invi

            with self.assertRaises(OutOfBandManagerError) as context:
                result = await self.manager.receive_invitation(
                    mock_oob_invi, use_existing_connection=True
                )
                assert "auto_present setting in configuration is" in str(
                    context.exception
                )

    async def test_req_attach_presentation_existing_conn_auto_present_no_pres_msg(self):
        self.session.context.update_settings({"public_invites": True})
        self.session.context.update_settings(
            {"debug.auto_respond_presentation_request": True}
        )
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        exchange_rec = V10PresentationExchange()
        exchange_rec.auto_present = True
        exchange_rec.presentation_request = TestConfig.INDY_PROOF_REQ

        presentation_proposal = PresentationProposal(
            comment="Hello World", presentation_proposal=TestConfig.PRES_PREVIEW
        )
        exchange_rec.presentation_proposal_dict = presentation_proposal.serialize()

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager,
            "receive_invitation",
            autospec=True,
        ) as didx_mgr_receive_invitation, async_mock.patch.object(
            PresentationManager,
            "receive_request",
            autospec=True,
        ) as proof_mgr_receive_request, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report, async_mock.patch.object(
            PresentationManager,
            "create_presentation",
            autospec=True,
        ) as proof_mgr_create_presentation:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            proof_mgr_receive_request.return_value = exchange_rec
            proof_mgr_create_presentation.return_value = (exchange_rec, None)
            holder = async_mock.MagicMock(IndyHolder, autospec=True)
            get_creds = async_mock.CoroutineMock(
                return_value=(
                    {
                        "cred_info": {"referent": "dummy_reft"},
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    },
                )
            )
            holder.get_credentials_for_presentation_request_by_referent = get_creds
            holder.create_credential_request = async_mock.CoroutineMock(
                return_value=(
                    json.dumps(TestConfig.indy_cred_req),
                    json.dumps(TestConfig.cred_req_meta),
                )
            )
            self.session.context.injector.bind_instance(IndyHolder, holder)
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[AttachDecorator.deserialize(TestConfig.req_attach)],
            )

            inv_message_cls.deserialize.return_value = mock_oob_invi

            with self.assertRaises(OutOfBandManagerError) as context:
                result = await self.manager.receive_invitation(
                    mock_oob_invi, use_existing_connection=True
                )
                assert "No presentation for proof request nonce" in str(
                    context.exception
                )

    async def test_req_attach_presentation_existing_conn_auto_present_pres_msg(self):
        self.session.context.update_settings({"public_invites": True})
        self.session.context.update_settings(
            {"debug.auto_respond_presentation_request": True}
        )
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        exchange_rec = V10PresentationExchange()
        exchange_rec.auto_present = True
        exchange_rec.presentation_request = TestConfig.INDY_PROOF_REQ
        exchange_rec.presentation_proposal_dict = {}

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager,
            "receive_invitation",
            autospec=True,
        ) as didx_mgr_receive_invitation, async_mock.patch.object(
            PresentationManager,
            "receive_request",
            autospec=True,
        ) as proof_mgr_receive_request, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report, async_mock.patch.object(
            PresentationManager,
            "create_presentation",
            autospec=True,
        ) as proof_mgr_create_presentation:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            proof_mgr_receive_request.return_value = exchange_rec
            proof_mgr_create_presentation.return_value = (
                exchange_rec,
                Presentation(comment="this is test"),
            )
            holder = async_mock.MagicMock(IndyHolder, autospec=True)
            get_creds = async_mock.CoroutineMock(
                return_value=(
                    {
                        "cred_info": {"referent": "dummy_reft"},
                        "attrs": {
                            "player": "Richie Knucklez",
                            "screenCapture": "aW1hZ2luZSBhIHNjcmVlbiBjYXB0dXJl",
                            "highScore": "1234560",
                        },
                    },
                )
            )
            holder.get_credentials_for_presentation_request_by_referent = get_creds
            holder.create_credential_request = async_mock.CoroutineMock(
                return_value=(
                    json.dumps(TestConfig.indy_cred_req),
                    json.dumps(TestConfig.cred_req_meta),
                )
            )
            self.session.context.injector.bind_instance(IndyHolder, holder)
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[AttachDecorator.deserialize(TestConfig.req_attach)],
            )

            inv_message_cls.deserialize.return_value = mock_oob_invi

            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=True
            )
            assert result.get("comment") == "this is test"

    async def test_req_attach_presentation_catch_value_error(self):
        self.session.context.update_settings({"public_invites": True})
        self.session.context.update_settings(
            {"debug.auto_respond_presentation_request": True}
        )
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")
        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )

        exchange_rec = V10PresentationExchange()
        exchange_rec.auto_present = True
        exchange_rec.presentation_request = TestConfig.INDY_PROOF_REQ
        exchange_rec.presentation_proposal_dict = {}

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager,
            "receive_invitation",
            autospec=True,
        ) as didx_mgr_receive_invitation, async_mock.patch.object(
            PresentationManager,
            "receive_request",
            autospec=True,
        ) as proof_mgr_receive_request, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report, async_mock.patch.object(
            PresentationManager,
            "create_presentation",
            autospec=True,
        ) as proof_mgr_create_presentation, async_mock.patch.object(
            PresentationProposal,
            "deserialize",
            autospec=True,
        ) as present_proposal_deserialize:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            proof_mgr_receive_request.return_value = exchange_rec
            proof_mgr_create_presentation.return_value = (
                exchange_rec,
                Presentation(comment="this is test"),
            )
            present_proposal_deserialize.return_value = PresentationProposal()
            holder = async_mock.MagicMock(IndyHolder, autospec=True)
            get_creds = async_mock.CoroutineMock(return_value=())
            holder.get_credentials_for_presentation_request_by_referent = get_creds
            holder.create_credential_request = async_mock.CoroutineMock(
                return_value=(
                    json.dumps(TestConfig.indy_cred_req),
                    json.dumps(TestConfig.cred_req_meta),
                )
            )
            self.session.context.injector.bind_instance(IndyHolder, holder)
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[AttachDecorator.deserialize(TestConfig.req_attach)],
            )

            inv_message_cls.deserialize.return_value = mock_oob_invi
            result = await self.manager.receive_invitation(
                mock_oob_invi, use_existing_connection=True
            )
            assert result is None

    async def test_req_attach_presentation_cred_offer(self):
        self.session.context.update_settings({"public_invites": True})
        self.session.context.update_settings(
            {"debug.auto_respond_presentation_request": True}
        )
        test_exist_conn = ConnRecord(
            my_did=TestConfig.test_did,
            their_did=TestConfig.test_target_did,
            their_public_did=TestConfig.test_target_did,
            invitation_msg_id="12345678-1234-5678-1234-567812345678",
            their_role=ConnRecord.Role.REQUESTER,
        )
        await test_exist_conn.save(self.session)
        await test_exist_conn.metadata_set(self.session, "reuse_msg_state", "initial")
        await test_exist_conn.metadata_set(self.session, "reuse_msg_id", "test_123")

        receipt = MessageReceipt(
            recipient_did=TestConfig.test_did,
            recipient_did_public=False,
            sender_did=TestConfig.test_target_did,
        )
        req_attach = deepcopy(TestConfig.req_attach)
        req_attach["data"]["json"]["@type"] = DIDCommPrefix.qualify_current(
            CREDENTIAL_OFFER
        )

        exchange_rec = V10PresentationExchange()
        exchange_rec.auto_present = True
        exchange_rec.presentation_request = TestConfig.INDY_PROOF_REQ
        exchange_rec.presentation_proposal_dict = {}

        with async_mock.patch.object(
            self.ledger, "get_key_for_did", async_mock.CoroutineMock()
        ) as mock_ledger_get_key_for_did, async_mock.patch.object(
            DIDXManager,
            "receive_invitation",
            autospec=True,
        ) as didx_mgr_receive_invitation, async_mock.patch.object(
            PresentationManager,
            "receive_request",
            autospec=True,
        ) as proof_mgr_receive_request, async_mock.patch(
            "aries_cloudagent.protocols.out_of_band.v1_0.manager.InvitationMessage",
            autospec=True,
        ) as inv_message_cls, async_mock.patch.object(
            OutOfBandManager,
            "fetch_connection_targets",
            autospec=True,
        ) as oob_mgr_fetch_conn, async_mock.patch.object(
            OutOfBandManager,
            "find_existing_connection",
            autospec=True,
        ) as oob_mgr_find_existing_conn, async_mock.patch.object(
            OutOfBandManager,
            "check_reuse_msg_state",
            autospec=True,
        ) as oob_mgr_check_reuse_state, async_mock.patch.object(
            OutOfBandManager,
            "create_handshake_reuse_message",
            autospec=True,
        ) as oob_mgr_create_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_message",
            autospec=True,
        ) as oob_mgr_receive_reuse_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_reuse_accepted_message",
            autospec=True,
        ) as oob_mgr_receive_accept_msg, async_mock.patch.object(
            OutOfBandManager,
            "receive_problem_report",
            autospec=True,
        ) as oob_mgr_receive_problem_report, async_mock.patch.object(
            PresentationManager,
            "create_presentation",
            autospec=True,
        ) as proof_mgr_create_presentation:
            mock_ledger_get_key_for_did.return_value = TestConfig.test_verkey
            oob_mgr_find_existing_conn.return_value = test_exist_conn
            proof_mgr_create_presentation.return_value = (
                exchange_rec,
                Presentation(comment="this is test"),
            )
            mock_oob_invi = async_mock.MagicMock(
                handshake_protocols=[
                    pfx.qualify(HSProto.RFC23.name) for pfx in DIDCommPrefix
                ],
                service_dids=[TestConfig.test_target_did],
                service_blocks=[],
                request_attach=[AttachDecorator.deserialize(req_attach)],
            )
            inv_message_cls.deserialize.return_value = mock_oob_invi
            with self.assertRaises(OutOfBandManagerError) as context:
                result = await self.manager.receive_invitation(
                    mock_oob_invi, use_existing_connection=True
                )
                assert "Unsupported request~attach type," in str(context.exception)



"""Credential exchange admin routes."""

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from json.decoder import JSONDecodeError
from marshmallow import fields, validate

from ....admin.request_context import AdminRequestContext
from ....connections.models.conn_record import ConnRecord
from ....core.profile import Profile
from ....indy.issuer import IndyIssuerError
from ....ledger.error import LedgerError
from ....messaging.credential_definitions.util import CRED_DEF_TAGS
from ....messaging.models.base import BaseModelError, OpenAPISchema
from ....messaging.valid import (
    INDY_CRED_DEF_ID,
    INDY_DID,
    INDY_SCHEMA_ID,
    INDY_VERSION,
    UUIDFour,
    UUID4,
)
from ....storage.error import StorageError, StorageNotFoundError
from ....wallet.base import BaseWallet
from ....wallet.error import WalletError
from ....utils.outofband import serialize_outofband
from ....utils.tracing import trace_event, get_timer, AdminAPIMessageTracingSchema

from ...problem_report.v1_0 import internal_error
from ...problem_report.v1_0.message import ProblemReport

from .manager import CredentialManager, CredentialManagerError
from .message_types import SPEC_URI
from .messages.credential_proposal import CredentialProposal
from .messages.credential_offer import CredentialOfferSchema
from .messages.inner.credential_preview import (
    CredentialPreview,
    CredentialPreviewSchema,
)
from .models.credential_exchange import (
    V10CredentialExchange,
    V10CredentialExchangeSchema,
)


class IssueCredentialModuleResponseSchema(OpenAPISchema):
    """Response schema for Issue Credential Module."""


class V10CredentialExchangeListQueryStringSchema(OpenAPISchema):
    """Parameters and validators for credential exchange list query."""

    connection_id = fields.UUID(
        description="Connection identifier",
        required=False,
        example=UUIDFour.EXAMPLE,  # typically but not necessarily a UUID4
    )
    thread_id = fields.UUID(
        description="Thread identifier",
        required=False,
        example=UUIDFour.EXAMPLE,  # typically but not necessarily a UUID4
    )
    role = fields.Str(
        description="Role assigned in credential exchange",
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10CredentialExchange, m)
                for m in vars(V10CredentialExchange)
                if m.startswith("ROLE_")
            ]
        ),
    )
    state = fields.Str(
        description="Credential exchange state",
        required=False,
        validate=validate.OneOf(
            [
                getattr(V10CredentialExchange, m)
                for m in vars(V10CredentialExchange)
                if m.startswith("STATE_")
            ]
        ),
    )


class V10CredentialExchangeListResultSchema(OpenAPISchema):
    """Result schema for Aries#0036 v1.0 credential exchange query."""

    results = fields.List(
        fields.Nested(V10CredentialExchangeSchema),
        description="Aries#0036 v1.0 credential exchange records",
    )


class V10CredentialStoreRequestSchema(OpenAPISchema):
    """Request schema for sending a credential store admin message."""

    credential_id = fields.Str(required=False)


class V10CredentialCreateSchema(AdminAPIMessageTracingSchema):
    """Base class for request schema for sending credential proposal admin message."""

    cred_def_id = fields.Str(
        description="Credential definition identifier",
        required=False,
        **INDY_CRED_DEF_ID,
    )
    schema_id = fields.Str(
        description="Schema identifier", required=False, **INDY_SCHEMA_ID
    )
    schema_issuer_did = fields.Str(
        description="Schema issuer DID", required=False, **INDY_DID
    )
    schema_name = fields.Str(
        description="Schema name", required=False, example="preferences"
    )
    schema_version = fields.Str(
        description="Schema version", required=False, **INDY_VERSION
    )
    issuer_did = fields.Str(
        description="Credential issuer DID", required=False, **INDY_DID
    )
    auto_remove = fields.Bool(
        description=(
            "Whether to remove the credential exchange record on completion "
            "(overrides --preserve-exchange-records configuration setting)"
        ),
        required=False,
    )
    comment = fields.Str(
        description="Human-readable comment", required=False, allow_none=True
    )
    trace = fields.Bool(
        description="Whether to trace event (default false)",
        required=False,
        example=False,
    )
    credential_proposal = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialProposalRequestSchemaBase(AdminAPIMessageTracingSchema):
    """Base class for request schema for sending credential proposal admin message."""

    connection_id = fields.UUID(
        description="Connection identifier",
        required=True,
        example=UUIDFour.EXAMPLE,  # typically but not necessarily a UUID4
    )
    cred_def_id = fields.Str(
        description="Credential definition identifier",
        required=False,
        **INDY_CRED_DEF_ID,
    )
    schema_id = fields.Str(
        description="Schema identifier", required=False, **INDY_SCHEMA_ID
    )
    schema_issuer_did = fields.Str(
        description="Schema issuer DID", required=False, **INDY_DID
    )
    schema_name = fields.Str(
        description="Schema name", required=False, example="preferences"
    )
    schema_version = fields.Str(
        description="Schema version", required=False, **INDY_VERSION
    )
    issuer_did = fields.Str(
        description="Credential issuer DID", required=False, **INDY_DID
    )
    auto_remove = fields.Bool(
        description=(
            "Whether to remove the credential exchange record on completion "
            "(overrides --preserve-exchange-records configuration setting)"
        ),
        required=False,
    )
    comment = fields.Str(
        description="Human-readable comment", required=False, allow_none=True
    )
    trace = fields.Bool(
        description="Whether to trace event (default false)",
        required=False,
        example=False,
    )


class V10CredentialProposalRequestOptSchema(V10CredentialProposalRequestSchemaBase):
    """Request schema for sending credential proposal on optional proposal preview."""

    credential_proposal = fields.Nested(CredentialPreviewSchema, required=False)


class V10CredentialProposalRequestMandSchema(V10CredentialProposalRequestSchemaBase):
    """Request schema for sending credential proposal on mandatory proposal preview."""

    credential_proposal = fields.Nested(CredentialPreviewSchema, required=True)


class V10CredentialOfferRequestSchema(AdminAPIMessageTracingSchema):
    """Request schema for sending credential offer admin message."""

    connection_id = fields.UUID(
        description="Connection identifier",
        required=True,
        example=UUIDFour.EXAMPLE,  # typically but not necessarily a UUID4
    )
    cred_def_id = fields.Str(
        description="Credential definition identifier",
        required=True,
        **INDY_CRED_DEF_ID,
    )
    auto_issue = fields.Bool(
        description=(
            "Whether to respond automatically to credential requests, creating "
            "and issuing requested credentials"
        ),
        required=False,
    )
    auto_remove = fields.Bool(
        description=(
            "Whether to remove the credential exchange record on completion "
            "(overrides --preserve-exchange-records configuration setting)"
        ),
        required=False,
        default=True,
    )
    comment = fields.Str(
        description="Human-readable comment", required=False, allow_none=True
    )
    credential_preview = fields.Nested(CredentialPreviewSchema, required=True)
    trace = fields.Bool(
        description="Whether to trace event (default false)",
        required=False,
        example=False,
    )


class V10CredentialIssueRequestSchema(OpenAPISchema):
    """Request schema for sending credential issue admin message."""

    comment = fields.Str(
        description="Human-readable comment", required=False, allow_none=True
    )


class V10CredentialProblemReportRequestSchema(OpenAPISchema):
    """Request schema for sending problem report."""

    explain_ltxt = fields.Str(required=True)


class CredIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential id."""

    credential_id = fields.Str(
        description="Credential identifier", required=True, example=UUIDFour.EXAMPLE
    )


class CredExIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking credential exchange id."""

    cred_ex_id = fields.Str(
        description="Credential exchange identifier", required=True, **UUID4
    )


@docs(
    tags=["issue-credential v1.0"],
    summary="Fetch all credential exchange records",
)
@querystring_schema(V10CredentialExchangeListQueryStringSchema)
@response_schema(V10CredentialExchangeListResultSchema(), 200, description="")
async def credential_exchange_list(request: web.BaseRequest):
    """
    Request handler for searching credential exchange records.

    Args:
        request: aiohttp request object

    Returns:
        The connection list response

    """
    context: AdminRequestContext = request["context"]
    tag_filter = {}
    if "thread_id" in request.query and request.query["thread_id"] != "":
        tag_filter["thread_id"] = request.query["thread_id"]
    post_filter = {
        k: request.query[k]
        for k in ("connection_id", "role", "state")
        if request.query.get(k, "") != ""
    }

    try:
        async with context.session() as session:
            records = await V10CredentialExchange.query(
                session=session,
                tag_filter=tag_filter,
                post_filter_positive=post_filter,
            )
        results = [record.serialize() for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(
    tags=["issue-credential v1.0"],
    summary="Fetch a single credential exchange record",
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_retrieve(request: web.BaseRequest):
    """
    Request handler for fetching single credential exchange record.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    try:
        async with context.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
        result = cred_ex_record.serialize()
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except (BaseModelError, StorageError) as err:
        await internal_error(err, web.HTTPBadRequest, cred_ex_record, outbound_handler)

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential, automating entire flow",
)
@request_schema(V10CredentialCreateSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_create(request: web.BaseRequest):
    """
    Request handler for creating a credential from attr values.

    The internal credential record will be created without the credential
    being sent to any connection. This can be used in conjunction with
    the `oob` protocols to bind messages to an out of band message.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]

    body = await request.json()

    comment = body.get("comment")
    preview_spec = body.get("credential_proposal")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="credential_proposal must be provided")
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")

    try:
        preview = CredentialPreview.deserialize(preview_spec)

        credential_proposal = CredentialProposal(
            comment=comment,
            credential_proposal=preview,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )
        credential_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            credential_proposal,
            outcome="credential_exchange_create.START",
        )

        credential_manager = CredentialManager(context.profile)

        (
            credential_exchange_record,
            credential_offer_message,
        ) = await credential_manager.prepare_send(
            None,
            credential_proposal=credential_proposal,
            auto_remove=auto_remove,
        )
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_create.END",
        perf_counter=r_time,
    )

    return web.json_response(credential_exchange_record.serialize())


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential, automating entire flow",
)
@request_schema(V10CredentialProposalRequestMandSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_send(request: web.BaseRequest):
    """
    Request handler for sending credential from issuer to holder from attr values.

    If both issuer and holder are configured for automatic responses, the operation
    ultimately results in credential issue; otherwise, the result waits on the first
    response not automated; the credential exchange record retains state regardless.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    comment = body.get("comment")
    connection_id = body.get("connection_id")
    preview_spec = body.get("credential_proposal")
    if not preview_spec:
        raise web.HTTPBadRequest(reason="credential_proposal must be provided")
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")

    connection_record = None
    cred_ex_record = None
    try:
        preview = CredentialPreview.deserialize(preview_spec)
        async with context.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_proposal = CredentialProposal(
            comment=comment,
            credential_proposal=preview,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )
        credential_proposal.assign_trace_decorator(
            context.settings,
            trace_msg,
        )

        trace_event(
            context.settings,
            credential_proposal,
            outcome="credential_exchange_send.START",
        )

        credential_manager = CredentialManager(context.profile)
        (
            cred_ex_record,
            credential_offer_message,
        ) = await credential_manager.prepare_send(
            connection_id,
            credential_proposal=credential_proposal,
            auto_remove=auto_remove,
        )
        result = cred_ex_record.serialize()

    except (StorageError, BaseModelError, CredentialManagerError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(
        credential_offer_message, connection_id=cred_ex_record.connection_id
    )

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send issuer a credential proposal",
)
@request_schema(V10CredentialProposalRequestOptSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_send_proposal(request: web.BaseRequest):
    """
    Request handler for sending credential proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    comment = body.get("comment")
    preview_spec = body.get("credential_proposal")
    auto_remove = body.get("auto_remove")
    trace_msg = body.get("trace")

    connection_record = None
    cred_ex_record = None
    try:
        preview = CredentialPreview.deserialize(preview_spec) if preview_spec else None
        async with context.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(context.profile)
        cred_ex_record = await credential_manager.create_proposal(
            connection_id,
            comment=comment,
            credential_preview=preview,
            auto_remove=auto_remove,
            trace=trace_msg,
            **{t: body.get(t) for t in CRED_DEF_TAGS if body.get(t)},
        )

        credential_proposal = CredentialProposal.deserialize(
            cred_ex_record.credential_proposal_dict
        )
        result = cred_ex_record.serialize()

    except (BaseModelError, StorageError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(
        credential_proposal,
        connection_id=connection_id,
    )

    trace_event(
        context.settings,
        credential_proposal,
        outcome="credential_exchange_send_proposal.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


async def _create_free_offer(
    profile: Profile,
    cred_def_id: str,
    connection_id: str = None,
    auto_issue: bool = False,
    auto_remove: bool = False,
    preview_spec: dict = None,
    comment: str = None,
    trace_msg: bool = None,
):
    """Create a credential offer and related exchange record."""

    credential_preview = CredentialPreview.deserialize(preview_spec)
    credential_proposal = CredentialProposal(
        comment=comment,
        credential_proposal=credential_preview,
        cred_def_id=cred_def_id,
    )
    credential_proposal.assign_trace_decorator(
        profile.settings,
        trace_msg,
    )
    credential_proposal_dict = credential_proposal.serialize()

    cred_ex_record = V10CredentialExchange(
        connection_id=connection_id,
        initiator=V10CredentialExchange.INITIATOR_SELF,
        role=V10CredentialExchange.ROLE_ISSUER,
        credential_definition_id=cred_def_id,
        credential_proposal_dict=credential_proposal_dict,
        auto_issue=auto_issue,
        auto_remove=auto_remove,
        trace=trace_msg,
    )

    credential_manager = CredentialManager(profile)

    (
        cred_ex_record,
        credential_offer_message,
    ) = await credential_manager.create_offer(cred_ex_record, comment=comment)

    return (cred_ex_record, credential_offer_message)


@docs(
    tags=["issue-credential v1.0"],
    summary="Create a credential offer, independent of any proposal",
)
@request_schema(V10CredentialOfferRequestSchema())
@response_schema(CredentialOfferSchema(), 200, description="")
async def credential_exchange_create_free_offer(request: web.BaseRequest):
    """
    Request handler for creating free credential offer.

    Unlike with `send-offer`, this credential exchange is not tied to a specific
    connection. It must be dispatched out-of-band by the controller.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    cred_def_id = body.get("cred_def_id")
    if not cred_def_id:
        raise web.HTTPBadRequest(reason="cred_def_id is required")

    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )
    auto_remove = body.get("auto_remove")
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    if not preview_spec:
        raise web.HTTPBadRequest(reason=("Missing credential_preview"))

    connection_id = body.get("connection_id")
    trace_msg = body.get("trace")

    async with context.session() as session:
        wallet = session.inject(BaseWallet)
        if connection_id:
            try:
                connection_record = await ConnRecord.retrieve_by_id(
                    session, connection_id
                )
                conn_did = await wallet.get_local_did(connection_record.my_did)
            except (WalletError, StorageError) as err:
                raise web.HTTPBadRequest(reason=err.roll_up) from err
        else:
            conn_did = await wallet.get_public_did()
            if not conn_did:
                raise web.HTTPBadRequest(reason="Wallet has no public DID")
            connection_id = None

        did_info = await wallet.get_public_did()
        del wallet

    endpoint = did_info.metadata.get(
        "endpoint", context.settings.get("default_endpoint")
    )
    if not endpoint:
        raise web.HTTPBadRequest(reason="An endpoint for the public DID is required")

    cred_ex_record = None
    try:
        (cred_ex_record, credential_offer_message) = await _create_free_offer(
            context.profile,
            cred_def_id,
            connection_id,
            auto_issue,
            auto_remove,
            preview_spec,
            comment,
            trace_msg,
        )

        trace_event(
            context.settings,
            credential_offer_message,
            outcome="credential_exchange_create_free_offer.END",
            perf_counter=r_time,
        )

        oob_url = serialize_outofband(credential_offer_message, conn_did, endpoint)
        result = cred_ex_record.serialize()

    except (BaseModelError, CredentialManagerError, LedgerError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    response = {"record": result, "oob_url": oob_url}
    return web.json_response(response)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential offer, independent of any proposal",
)
@request_schema(V10CredentialOfferRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_send_free_offer(request: web.BaseRequest):
    """
    Request handler for sending free credential offer.

    An issuer initiates a such a credential offer, free from any
    holder-initiated corresponding credential proposal with preview.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    body = await request.json()

    connection_id = body.get("connection_id")
    cred_def_id = body.get("cred_def_id")
    if not cred_def_id:
        raise web.HTTPBadRequest(reason="cred_def_id is required")

    auto_issue = body.get(
        "auto_issue", context.settings.get("debug.auto_respond_credential_request")
    )

    auto_remove = body.get("auto_remove")
    comment = body.get("comment")
    preview_spec = body.get("credential_preview")
    if not preview_spec:
        raise web.HTTPBadRequest(reason=("Missing credential_preview"))
    trace_msg = body.get("trace")

    cred_ex_record = None
    connection_record = None
    try:
        async with context.session() as session:
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        (cred_ex_record, credential_offer_message,) = await _create_free_offer(
            context.profile,
            cred_def_id,
            connection_id,
            auto_issue,
            auto_remove,
            preview_spec,
            comment,
            trace_msg,
        )
        result = cred_ex_record.serialize()

    except (
        StorageNotFoundError,
        BaseModelError,
        CredentialManagerError,
        LedgerError,
    ) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(credential_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send_free_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential offer in reference to a proposal with preview",
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_send_bound_offer(request: web.BaseRequest):
    """
    Request handler for sending bound credential offer.

    A holder initiates this sequence with a credential proposal; this message
    responds with an offer bound to the proposal.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    connection_record = None
    try:
        async with context.session() as session:
            try:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    session, credential_exchange_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            connection_id = cred_ex_record.connection_id
            if cred_ex_record.state != (
                V10CredentialExchange.STATE_PROPOSAL_RECEIVED
            ):  # check state here: manager call creates free offers too
                raise CredentialManagerError(
                    f"Credential exchange {cred_ex_record.credential_exchange_id} "
                    f"in {cred_ex_record.state} state "
                    f"(must be {V10CredentialExchange.STATE_PROPOSAL_RECEIVED})"
                )

            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(context.profile)
        (
            cred_ex_record,
            credential_offer_message,
        ) = await credential_manager.create_offer(cred_ex_record, comment=None)

        result = cred_ex_record.serialize()

    except (StorageError, BaseModelError, CredentialManagerError, LedgerError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(credential_offer_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_offer_message,
        outcome="credential_exchange_send_bound_offer.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send issuer a credential request",
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_send_request(request: web.BaseRequest):
    """
    Request handler for sending credential request.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    connection_record = None
    try:
        async with context.session() as session:
            try:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    session, credential_exchange_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err
            connection_id = cred_ex_record.connection_id

            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(context.profile)
        (
            cred_ex_record,
            credential_request_message,
        ) = await credential_manager.create_request(
            cred_ex_record, connection_record.my_did
        )

        result = cred_ex_record.serialize()

    except (StorageError, CredentialManagerError, BaseModelError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(credential_request_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_request_message,
        outcome="credential_exchange_send_request.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Send holder a credential",
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialIssueRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_issue(request: web.BaseRequest):
    """
    Request handler for sending credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    body = await request.json()
    comment = body.get("comment")

    credential_exchange_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    connection_record = None
    try:
        async with context.session() as session:
            try:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    session, credential_exchange_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err
            connection_id = cred_ex_record.connection_id

            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(context.profile)
        (
            cred_ex_record,
            credential_issue_message,
        ) = await credential_manager.issue_credential(cred_ex_record, comment=comment)

        result = cred_ex_record.serialize()

    except (
        BaseModelError,
        CredentialManagerError,
        IndyIssuerError,
        StorageError,
    ) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(credential_issue_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_issue_message,
        outcome="credential_exchange_issue.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Store a received credential",
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialStoreRequestSchema())
@response_schema(V10CredentialExchangeSchema(), 200, description="")
async def credential_exchange_store(request: web.BaseRequest):
    """
    Request handler for storing credential.

    Args:
        request: aiohttp request object

    Returns:
        The credential exchange record

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    try:
        body = await request.json() or {}
        credential_id = body.get("credential_id")
    except JSONDecodeError:
        credential_id = None

    credential_exchange_id = request.match_info["cred_ex_id"]

    cred_ex_record = None
    connection_record = None
    try:
        async with context.session() as session:
            try:
                cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                    session, credential_exchange_id
                )
            except StorageNotFoundError as err:
                raise web.HTTPNotFound(reason=err.roll_up) from err

            connection_id = cred_ex_record.connection_id
            connection_record = await ConnRecord.retrieve_by_id(session, connection_id)
            if not connection_record.is_ready:
                raise web.HTTPForbidden(reason=f"Connection {connection_id} not ready")

        credential_manager = CredentialManager(context.profile)
        (
            cred_ex_record,
            credential_stored_message,
        ) = await credential_manager.store_credential(cred_ex_record, credential_id)

        result = cred_ex_record.serialize()

    except (StorageError, CredentialManagerError, BaseModelError) as err:
        await internal_error(
            err,
            web.HTTPBadRequest,
            cred_ex_record or connection_record,
            outbound_handler,
        )

    await outbound_handler(credential_stored_message, connection_id=connection_id)

    trace_event(
        context.settings,
        credential_stored_message,
        outcome="credential_exchange_store.END",
        perf_counter=r_time,
    )

    return web.json_response(result)


@docs(
    tags=["issue-credential v1.0"],
    summary="Remove an existing credential exchange record",
)
@match_info_schema(CredExIdMatchInfoSchema())
@response_schema(IssueCredentialModuleResponseSchema(), 200, description="")
async def credential_exchange_remove(request: web.BaseRequest):
    """
    Request handler for removing a credential exchange record.

    Args:
        request: aiohttp request object

    """
    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    cred_ex_record = None
    try:
        async with context.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
            await cred_ex_record.delete_record(session)
    except StorageNotFoundError as err:
        await internal_error(err, web.HTTPNotFound, cred_ex_record, outbound_handler)
    except StorageError as err:
        await internal_error(err, web.HTTPBadRequest, cred_ex_record, outbound_handler)

    return web.json_response({})


@docs(
    tags=["issue-credential v1.0"],
    summary="Send a problem report for credential exchange",
)
@match_info_schema(CredExIdMatchInfoSchema())
@request_schema(V10CredentialProblemReportRequestSchema())
@response_schema(IssueCredentialModuleResponseSchema(), 200, description="")
async def credential_exchange_problem_report(request: web.BaseRequest):
    """
    Request handler for sending problem report.

    Args:
        request: aiohttp request object

    """
    r_time = get_timer()

    context: AdminRequestContext = request["context"]
    outbound_handler = request["outbound_message_router"]

    credential_exchange_id = request.match_info["cred_ex_id"]
    body = await request.json()

    try:
        async with await context.session() as session:
            cred_ex_record = await V10CredentialExchange.retrieve_by_id(
                session, credential_exchange_id
            )
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err

    error_result = ProblemReport(explain_ltxt=body["explain_ltxt"])
    error_result.assign_thread_id(cred_ex_record.thread_id)

    await outbound_handler(error_result, connection_id=cred_ex_record.connection_id)

    trace_event(
        context.settings,
        error_result,
        outcome="credential_exchange_problem_report.END",
        perf_counter=r_time,
    )

    return web.json_response({})


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get(
                "/issue-credential/records", credential_exchange_list, allow_head=False
            ),
            web.get(
                "/issue-credential/records/{cred_ex_id}",
                credential_exchange_retrieve,
                allow_head=False,
            ),
            web.post("/issue-credential/create", credential_exchange_create),
            web.post("/issue-credential/send", credential_exchange_send),
            web.post(
                "/issue-credential/send-proposal", credential_exchange_send_proposal
            ),
            web.post(
                "/issue-credential/send-offer", credential_exchange_send_free_offer
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/send-offer",
                credential_exchange_send_bound_offer,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/send-request",
                credential_exchange_send_request,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/issue",
                credential_exchange_issue,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/store",
                credential_exchange_store,
            ),
            web.post(
                "/issue-credential/records/{cred_ex_id}/problem-report",
                credential_exchange_problem_report,
            ),
            web.delete(
                "/issue-credential/records/{cred_ex_id}",
                credential_exchange_remove,
            ),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {
            "name": "issue-credential v1.0",
            "description": "Credential issue v1.0",
            "externalDocs": {"description": "Specification", "url": SPEC_URI},
        }
    )



"""The localization decorator (~l10n) for message localization information."""

from typing import Sequence

from marshmallow import EXCLUDE, fields

from ..models.base import BaseModel, BaseModelSchema


class LocalizationDecorator(BaseModel):
    """Class representing the localization decorator."""

    class Meta:
        """LocalizationDecorator metadata."""

        schema_class = "LocalizationDecoratorSchema"

    def __init__(
        self,
        *,
        locale: str = None,
        localizable: Sequence[str] = None,
        catalogs: Sequence[str] = None,
    ):
        """
        Initialize a LocalizationDecorator instance.

        Args:
            locale: The locale of this message
            localizable: The fields which may be localized
            catalogs: A list of URLs for localization resources

        """
        super().__init__()
        self.locale = locale
        self.localizable = list(localizable) if localizable else []
        self.catalogs = list(catalogs) if catalogs else []


class LocalizationDecoratorSchema(BaseModelSchema):
    """Localization decorator schema used in serialization/deserialization."""

    class Meta:
        """LocalizationDecoratorSchema metadata."""

        model_class = LocalizationDecorator
        unknown = EXCLUDE

    locale = fields.Str(
        required=True,
        description="Locale specifier",
        example="en-CA",
    )
    localizable = fields.List(
        fields.Str(
            description="Localizable field",
            example="note",
        ),
        required=False,
        description="List of localizable fields",
    )
    catalogs = fields.List(
        fields.Str(
            description="",
            example="https://192.168.56.111/my-project/catalog.json",
        ),
        required=False,
        description="List of message catalog URIs",
    )



"""Message type identifiers for problem reports."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/hyperledger/aries-rfcs/tree/"
    "89d14c15ab35b667e7a9d04fe42d4d48b10468cf/features/0035-report-problem"
)

# Message types
PROBLEM_REPORT = "notification/1.0/problem-report"

PROTOCOL_PACKAGE = "aries_cloudagent.protocols.problem_report.v1_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {PROBLEM_REPORT: f"{PROTOCOL_PACKAGE}.message.ProblemReport"}
)



import pytest

from unittest import TestCase

from ..base import SettingsError
from ..settings import Settings


class TestSettings(TestCase):
    def setUp(self):
        self.test_key = "TEST"
        self.test_value = "VALUE"
        self.test_settings = {self.test_key: self.test_value}
        self.test_instance = Settings(self.test_settings)

    def test_settings_init(self):
        """Test settings initialization."""
        for key in self.test_settings:
            assert key in self.test_instance
            assert self.test_instance[key] == self.test_settings[key]
            assert (
                self.test_instance.get_value(self.test_key) == self.test_settings[key]
            )
        with self.assertRaises(KeyError):
            self.test_instance["MISSING"]
        assert len(self.test_instance) == 1
        assert len(self.test_instance.copy()) == 1

    def test_get_formats(self):
        """Test retrieval with formatting."""
        assert "Settings" in str(self.test_instance)
        with pytest.raises(TypeError):
            self.test_instance[0]  # cover wrong type
        self.test_instance["BOOL"] = "true"
        assert self.test_instance.get_bool("BOOL") is True
        self.test_instance["BOOL"] = "false"
        assert self.test_instance.get_bool("BOOL") is False
        self.test_instance["INT"] = "5"
        assert self.test_instance.get_int("INT") is 5
        assert self.test_instance.get_str("INT") == "5"
        with self.assertRaises(TypeError):
            self.test_instance[None] = 1
        with self.assertRaises(ValueError):
            self.test_instance[""] = 1

    def test_remove(self):
        """Test value removal."""
        del self.test_instance[self.test_key]
        with self.assertRaises(KeyError):
            self.test_instance[self.test_key]
        self.test_instance[self.test_key] = self.test_value
        self.test_instance.clear_value(self.test_key)
        with self.assertRaises(KeyError):
            self.test_instance[self.test_key]

    def test_set_default(self):
        """Test default value."""
        self.test_instance.set_default(self.test_key, "RANDOM")
        assert self.test_instance[self.test_key] == self.test_value
        self.test_instance.set_default("BOOL", "True")
        assert self.test_instance["BOOL"] == "True"



"""Represents a connection response message."""

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema

from ..message_types import CONNECTION_RESPONSE, PROTOCOL_PACKAGE
from ..models.connection_detail import ConnectionDetail, ConnectionDetailSchema

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers."
    "connection_response_handler.ConnectionResponseHandler"
)


class ConnectionResponse(AgentMessage):
    """Class representing a connection response."""

    class Meta:
        """Metadata for a connection response."""

        handler_class = HANDLER_CLASS
        schema_class = "ConnectionResponseSchema"
        message_type = CONNECTION_RESPONSE

    def __init__(self, *, connection: ConnectionDetail = None, **kwargs):
        """
        Initialize connection response object.

        Args:
            connection: Connection details object

        """
        super().__init__(**kwargs)
        self.connection = connection


class ConnectionResponseSchema(AgentMessageSchema):
    """Connection response schema class."""

    class Meta:
        """Connection response schema metadata."""

        model_class = ConnectionResponse
        signed_fields = ("connection",)
        unknown = EXCLUDE

    connection = fields.Nested(ConnectionDetailSchema, required=True)



"""Base injection context builder classes."""

from abc import ABC, abstractmethod
from typing import Mapping

from .injection_context import InjectionContext
from .settings import Settings


class ContextBuilder(ABC):
    """Base injection context builder class."""

    def __init__(self, settings: Mapping[str, object] = None):
        """
        Initialize an instance of the context builder.

        Args:
            settings: Mapping of configuration settings

        """
        self.settings = Settings(settings)

    @abstractmethod
    async def build_context(self) -> InjectionContext:
        """Build the base injection context."""

    def update_settings(self, settings: Mapping[str, object]):
        """Update the context builder with additional settings."""
        if settings:
            self.settings = self.settings.extend(settings)



"""
Admin request context class.

A request context provided by the admin server to admin route handlers.
"""

from typing import Mapping, Optional, Type

from ..core.profile import Profile, ProfileSession
from ..config.injector import Injector, InjectionError, InjectType
from ..config.injection_context import InjectionContext
from ..config.settings import Settings
from ..utils.classloader import DeferLoad

IN_MEM = DeferLoad("aries_cloudagent.core.in_memory.InMemoryProfile")


class AdminRequestContext:
    """Context established by the Conductor and passed into message handlers."""

    def __init__(
        self,
        profile: Profile,
        *,
        context: InjectionContext = None,
        settings: Mapping[str, object] = None
    ):
        """Initialize an instance of AdminRequestContext."""
        self._context = (context or profile.context).start_scope("admin", settings)
        self._profile = profile

    @property
    def injector(self) -> Injector:
        """Accessor for the associated `Injector` instance."""
        return self._context.injector

    @property
    def profile(self) -> Profile:
        """Accessor for the associated `Profile` instance."""
        return self._profile

    @property
    def settings(self) -> Settings:
        """Accessor for the context settings."""
        return self._context.settings

    def session(self) -> ProfileSession:
        """Start a new interactive session with no transaction support requested."""
        return self.profile.session(self._context)

    def transaction(self) -> ProfileSession:
        """
        Start a new interactive session with commit and rollback support.

        If the current backend does not support transactions, then commit
        and rollback operations of the session will not have any effect.
        """
        return self.profile.transaction(self._context)

    def inject(
        self,
        base_cls: Type[InjectType],
        settings: Mapping[str, object] = None,
        *,
        required: bool = True
    ) -> Optional[InjectType]:
        """
        Get the provided instance of a given class identifier.

        Args:
            cls: The base class to retrieve an instance of
            settings: An optional mapping providing configuration to the provider

        Returns:
            An instance of the base class, or None

        """
        return self._context.inject(base_cls, settings, required=required)

    def update_settings(self, settings: Mapping[str, object]):
        """Update the current scope with additional settings."""
        self._context.update_settings(settings)

    @classmethod
    def test_context(
        cls, session_inject: dict = None, profile: Profile = None
    ) -> "AdminRequestContext":
        """Quickly set up a new admin request context for tests."""
        ctx = AdminRequestContext(profile or IN_MEM.resolved.test_profile())
        setattr(
            ctx, "session_inject", dict() if session_inject is None else session_inject
        )
        setattr(ctx, "session", ctx._test_session)
        return ctx

    def _test_session(self) -> ProfileSession:
        session = self.profile.session(self._context)

        def _inject(base_cls, required=True):
            if session._active and base_cls in self.session_inject:
                ret = self.session_inject[base_cls]
                if ret is None and required:
                    raise InjectionError(
                        "No instance provided for class: {}".format(base_cls.__name__)
                    )
                return ret
            return session._context.injector.inject(base_cls, required=required)

        setattr(session, "inject", _inject)
        return session

    def __repr__(self) -> str:
        """
        Provide a human readable representation of this object.

        Returns:
            A human readable representation of this object

        """
        skip = ("session",)
        items = (
            "{}={}".format(k, repr(v))
            for k, v in self.__dict__.items()
            if k not in skip
        )
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))



from asynctest import TestCase as AsyncTestCase

from ..base import BaseProvider, BaseInjector, BaseSettings, InjectionError
from ..injector import Injector
from ..provider import ClassProvider, CachedProvider


class MockProvider(BaseProvider):
    def __init__(self, value):
        self.settings = None
        self.injector = None
        self.value = value

    def provide(self, settings: BaseSettings, injector: BaseInjector):
        self.settings = settings
        self.injector = injector
        return self.value


class MockInstance:
    def __init__(self, value, **kwargs):
        self.opened = False
        self.value = value
        self.kwargs = kwargs

    def open(self):
        self.opened = True


class TestInjector(AsyncTestCase):
    def setUp(self):
        self.test_key = "TEST"
        self.test_value = "VALUE"
        self.test_settings = {self.test_key: self.test_value}
        self.test_instance = Injector(settings=self.test_settings)
        assert self.test_instance.__class__.__name__ in str(self.test_instance)

    def test_settings_init(self):
        """Test settings initialization."""
        for key in self.test_settings:
            assert key in self.test_instance.settings
            assert self.test_instance.settings[key] == self.test_settings[key]

    def test_inject_simple(self):
        """Test a basic injection."""
        assert self.test_instance.inject(str, required=False) is None
        with self.assertRaises(InjectionError):
            self.test_instance.inject(str)
        with self.assertRaises(ValueError):
            self.test_instance.bind_instance(str, None)
        self.test_instance.bind_instance(str, self.test_value)
        assert self.test_instance.inject(str) is self.test_value

    def test_inject_x(self):
        """Test injection failure on null base class."""
        with self.assertRaises(InjectionError):
            self.test_instance.inject(None)

    def test_inject_provider(self):
        """Test a provider injection."""
        mock_provider = MockProvider(self.test_value)

        with self.assertRaises(ValueError):
            self.test_instance.bind_provider(str, None)
        self.test_instance.bind_provider(str, mock_provider)
        assert self.test_instance.get_provider(str) is mock_provider

        override_settings = {self.test_key: "NEWVAL"}
        assert self.test_instance.inject(str, override_settings) is self.test_value
        assert mock_provider.settings[self.test_key] == override_settings[self.test_key]
        assert mock_provider.injector is self.test_instance

    def test_bad_provider(self):
        """Test empty and invalid provider results."""
        self.test_instance.bind_provider(str, MockProvider(None))
        with self.assertRaises(InjectionError):
            self.test_instance.inject(str)
        self.test_instance.inject(str, required=False)
        self.test_instance.bind_provider(str, MockProvider(1))
        self.test_instance.clear_binding(str)
        assert self.test_instance.get_provider(str) is None
        with self.assertRaises(InjectionError):
            self.test_instance.inject(str)

    def test_inject_class(self):
        """Test a provider class injection."""
        provider = ClassProvider(MockInstance, self.test_value, init_method="open")
        self.test_instance.bind_provider(MockInstance, provider)
        assert self.test_instance.get_provider(MockInstance) is provider
        instance = self.test_instance.inject(MockInstance)
        assert isinstance(instance, MockInstance)
        assert instance.value is self.test_value
        assert instance.opened

    def test_inject_class_name(self):
        """Test a provider class injection with a named class."""
        provider = ClassProvider("aries_cloudagent.config.settings.Settings")
        self.test_instance.bind_provider(BaseSettings, provider)
        instance = self.test_instance.inject(BaseSettings)
        assert (
            isinstance(instance, BaseSettings)
            and instance.__class__.__name__ == "Settings"
        )

    def test_inject_class_dependency(self):
        """Test a provider class injection with a dependency."""
        test_str = "TEST"
        test_int = 1
        self.test_instance.bind_instance(str, test_str)
        self.test_instance.bind_instance(int, test_int)
        provider = ClassProvider(
            MockInstance, ClassProvider.Inject(str), param=ClassProvider.Inject(int)
        )
        self.test_instance.bind_provider(object, provider)
        instance = self.test_instance.inject(object)
        assert instance.value is test_str
        assert instance.kwargs["param"] is test_int

        self.test_instance.clear_binding(int)
        self.test_instance.clear_binding(str)
        self.test_instance.bind_instance(str, test_int)
        with self.assertRaises(InjectionError):
            self.test_instance.inject(str)

    def test_inject_cached(self):
        """Test a provider class injection."""
        with self.assertRaises(ValueError):
            CachedProvider(None)
        provider = ClassProvider(MockInstance, self.test_value, init_method="open")
        cached = CachedProvider(provider)
        self.test_instance.bind_provider(MockInstance, cached)
        assert self.test_instance.get_provider(MockInstance) is cached
        i1 = self.test_instance.inject(MockInstance)
        i2 = self.test_instance.inject(MockInstance)
        assert i1 is i2

        provider = ClassProvider(MockInstance, self.test_value, init_method="open")
        self.test_instance.bind_provider(MockInstance, provider, cache=True)
        i1 = self.test_instance.inject(MockInstance)
        i2 = self.test_instance.inject(MockInstance)
        assert i1 is i2



from asynctest import (
    mock as async_mock,
    TestCase as AsyncTestCase,
)

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......transport.inbound.receipt import MessageReceipt

from ...messages.cred_ack import V20CredAck
from .. import cred_ack_handler as test_module


class TestCredentialAckHandler(AsyncTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context()
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = async_mock.MagicMock()

        with async_mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_credential_ack = (
                async_mock.CoroutineMock()
            )
            request_context.message = V20CredAck()
            request_context.connection_ready = True
            handler = test_module.V20CredAckHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_cred_mgr.assert_called_once_with(request_context.profile)
        mock_cred_mgr.return_value.receive_credential_ack.assert_called_once_with(
            request_context.message,
            request_context.connection_record.connection_id,
        )
        assert not responder.messages

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context()
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = async_mock.MagicMock()

        with async_mock.patch.object(
            test_module, "V20CredManager", autospec=True
        ) as mock_cred_mgr:
            mock_cred_mgr.return_value.receive_cred_ack = async_mock.CoroutineMock()
            request_context.message = V20CredAck()
            request_context.connection_ready = False
            handler = test_module.V20CredAckHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException):
                await handler.handle(request_context, responder)

        assert not responder.messages
